package seu.agent.BasicTask;

import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import rescuecore2.misc.Pair;
import rescuecore2.misc.geometry.Line2D;
import rescuecore2.misc.geometry.Point2D;
import rescuecore2.standard.entities.Area;
import rescuecore2.standard.entities.Blockade;
import rescuecore2.standard.entities.Edge;
import rescuecore2.standard.entities.Human;
import rescuecore2.standard.entities.Road;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardWorldModel;
import rescuecore2.worldmodel.EntityID;
import seu.agent.SEUAgent;
import seu.agent.command.seuCommand;
import seu.agent.command.seuMove;
import seu.agent.command.seuMoveXY;
import seu.utils.geo.GeoUtils;
import seu.utils.geo.Line;
import seu.utils.geo.Point;
import seu.utils.geo.Polygon;
import seu.utils.geo.Segment;

/**
 * 
 * @author Oliver, Fun
 * 
 */
public abstract class AbstractStuckTask extends
		BasicTask<SEUAgent<? extends StandardEntity>> {
	private static final int LEAST_MOVE_LEN = 2500;
	private static final double TRANSLATION_COEFFICIENT = 1;

	protected Human me;
	protected StandardWorldModel world;
	protected int ignoreTimestamp;
	SEUAgent<? extends StandardEntity> agent;

	public AbstractStuckTask(SEUAgent<? extends StandardEntity> agent) {
		super(agent);
		this.me = agent.getMeAsHuman();
		this.world = agent.getModel();
		this.ignoreTimestamp = agent.getStartTime();
		this.agent = agent;
	}

	@Override
	public abstract void taskDone();

	@Override
	public String toString() {
		return AbstractStuckTask.class.getName();
	}

	protected boolean isStuck(int checkStep) {
		if (!isAlwaysMove(checkStep)) {
			return false;
		}
		// FIXME 一些极端case
		if (!isAlwaysSamePlace(checkStep)) {
			return false;
		}
		if (!isNotMoveTooLong(checkStep)) {
			return false;
		}
		return true;
	}

	protected boolean isAlwaysMove(int checkStep) {
		int time = agent.getTimeStep();
		for (int i = 1; i <= checkStep; i++) {
			seuCommand ac = agent.getCommands().get(time - i);
			if ((ac instanceof seuMove) || (ac instanceof seuMoveXY)) {

			} else {
				return false;
			}
		}
		return true;
	}

	private boolean isAlwaysSamePlace(int checkStep) {
		int time = agent.getTimeStep();
		for (int i = 1; i < checkStep; i++) {
			EntityID info1 = agent.getLocation(time - i);
			EntityID info2 = agent.getLocation(time - i + 1);
			if (!info1.equals(info2))
				return false;

		}
		return true;
	}

	private boolean isNotMoveTooLong(int checkStep) {
		int time = agent.getTimeStep();
		int totals = 0;
		for (int i = 1; i < checkStep; i++) {
			EntityID info1 = agent.getLocation(time - i);
			EntityID info2 = agent.getLocation(time - i + 1);
			totals += agent.getModel().manhattanDistance(info1, info2);
		}
		if (totals > LEAST_MOVE_LEN) {
			return false;
		} else {
			return true;
		}
	}

	/**
	 * 用Blockade截取相邻边，得到中点
	 * 
	 * @param agent
	 * @param nextArea
	 * @return
	 */
	protected Pair<Integer, Integer> getTargetXY(EntityID nextArea) {
		StandardWorldModel world = agent.getModel();

		Area ahere = (Area) world.getEntity(agent.getMeAsHuman().getPosition());
		Area anext = (Area) world.getEntity(nextArea);

		Road where = null;
		Road wnext = null;

		if (ahere instanceof Road) {
			where = (Road) ahere;
		}
		if (anext instanceof Road) {
			wnext = (Road) anext;
		}

		Edge cedge = ahere.getEdgeTo(nextArea);
		if (cedge == null) {
			return null;
		}
		Segment seg = new Segment(cedge.getStartX(), cedge.getStartY(),
				cedge.getEndX(), cedge.getEndY());

		if (where != null) {
			for (Polygon p : agent.getModel().getExpBlockades(where)) {
				seg = GeoUtils.cutSegmentInsidePolygon(seg, p);
				if (seg == null) {
					return null;
				}
			}
		}

		if (wnext != null) {
			for (Polygon p : agent.getModel().getExpBlockades(wnext)) {
				seg = GeoUtils.cutSegmentInsidePolygon(seg, p);
				if (seg == null) {
					return null;
				}
			}
		}

		return new Pair<Integer, Integer>(seg.getMidX(), seg.getMidY());
	}

	/**
	 * 离散化相邻边，取离Blockade最远点
	 * 
	 * @param agent
	 * @param nextArea
	 * @return
	 */
	protected Pair<Integer, Integer> getTargetXY2(EntityID nextArea) {
		StandardWorldModel world = agent.getModel();

		Area ahere = (Area) world.getEntity(agent.getMeAsHuman().getPosition());
		Area anext = (Area) world.getEntity(nextArea);

		Road where = null;
		Road wnext = null;

		if (ahere instanceof Road) {
			where = (Road) ahere;
		}
		if (anext instanceof Road) {
			wnext = (Road) anext;
		}

		Edge cedge = ahere.getEdgeTo(nextArea);
		if (cedge == null) {
			return null;
		}
		Segment seg = new Segment(cedge.getStartX(), cedge.getStartY(),
				cedge.getEndX(), cedge.getEndY());

		List<Segment> dseg = GeoUtils.discretizeSegment(seg,
				(int) (agent.getAgentRadius() * 2 + 200));

		if (dseg.size() <= 0 || dseg.size() >= 25) { // 复杂度高
			return null;
		}

		Map<Point, Integer> pdmap = new HashMap<Point, Integer>();

		for (Segment next : dseg) {
			pdmap.put(next.getMidPoint(), Integer.MAX_VALUE);
		}
		if (where != null) {
			for (Blockade blockade : agent.getModel().getBlockades(where)) {
				Point bcenter = new Point(blockade.getX(), blockade.getY());
				for (Entry<Point, Integer> entry : pdmap.entrySet()) {
					int dist = (int) GeoUtils.dist(entry.getKey(), bcenter);
					if (dist < entry.getValue()) {
						entry.setValue(dist);
					}
				}
			}
		}

		if (wnext != null) {
			for (Blockade blockade : agent.getModel().getBlockades(wnext)) {
				Point bcenter = new Point(blockade.getX(), blockade.getY());
				for (Entry<Point, Integer> entry : pdmap.entrySet()) {
					int dist = (int) GeoUtils.dist(entry.getKey(), bcenter);
					if (dist < entry.getValue()) {
						entry.setValue(dist);
					}
				}
			}
		}

		int dist = 0;
		Point p = null;
		for (Entry<Point, Integer> entry : pdmap.entrySet()) {
			int tmp = entry.getValue();
			if (tmp > dist) {
				dist = tmp;
				p = entry.getKey();
			}
		}
		if (p == null) {
			return null;
		}
		return new Pair<Integer, Integer>(p.getX(), p.getY());
	}

	protected static Line getPassableLine(Line2D orignalLine,
			Collection<Blockade> blockades) {
		Set<Pair<Double, Double>> pieces = new HashSet<Pair<Double, Double>>();
		pieces.add(new Pair<Double, Double>(0d, 1d));
		for (Blockade b : blockades) {
			double minU = Double.MAX_VALUE;
			double maxU = Double.MIN_VALUE;
			int[] apex = b.getApexes();
			for (int i = 0; i < apex.length - 1; i += 2) {
				double u = getClosestPointRelativePostionOnSegment(orignalLine,
						apex[i], apex[i + 1]);
				if (minU > u) {
					minU = u;
				}
				if (maxU < u) {
					maxU = u;
				}
			}

			Set<Pair<Double, Double>> newPieces = new HashSet<Pair<Double, Double>>();
			for (Iterator<Pair<Double, Double>> it = pieces.iterator(); it
					.hasNext();) {
				Pair<Double, Double> piece = it.next();
				Double pieceMin = piece.first();
				Double pieceMax = piece.second();
				if (maxU <= pieceMin || minU >= pieceMax) {
					continue;
				}
				if (maxU <= pieceMax && minU >= pieceMin) {
					// split
					if (minU != pieceMin) {
						newPieces.add(new Pair<Double, Double>(pieceMin, minU));
					}
					if (maxU != pieceMax) {
						newPieces.add(new Pair<Double, Double>(maxU, pieceMax));
					}
				} else if (maxU >= pieceMax && minU <= pieceMin) {
					// cover
				} else {
					if (maxU < pieceMax) {
						// left shrink
						newPieces.add(new Pair<Double, Double>(maxU, pieceMax));
					} else if (minU > pieceMin) {
						// right shrink
						newPieces.add(new Pair<Double, Double>(pieceMin, minU));
					}
				}
				it.remove();
			}
			pieces.addAll(newPieces);
			
		}
		double maxD = Double.MIN_VALUE;
		Pair<Double, Double> thePiece = null;
		for (Pair<Double, Double> piece : pieces) {
			double d = piece.second() - piece.first();
			if (d > maxD) {
				maxD = d;
				thePiece = piece;
			}
		}

		if (thePiece != null) {
			double pieceLength = orignalLine.getDirection().getLength() * maxD;
			if (pieceLength > 500 * 2) {
				// translation away edgeLine center
				double dU = ((thePiece.first() + thePiece.second()) / 2 - 0.5)
						* TRANSLATION_COEFFICIENT;
				// log.info("Found the piece:" + thePiece + ", pieceLength:" +
				// pieceLength + " dU:"
				// + dU);
				Point2D p1 = orignalLine.getPoint(Math.min(
						Math.max(thePiece.first() + dU, 0), 0.99));
				Point2D p2 = orignalLine.getPoint(Math.max(
						Math.min(thePiece.second() + dU, 1), 0.01));
				return new Line(new Point((int) p1.getX(), (int) p1.getY()),
						new Point((int) p2.getX(), (int) p2.getY()));
			} else {
				 //log.info("piece is not long enough! " + thePiece +
				// ", pieceLength:" + pieceLength);
			}
		} else {
			// log.info("Not found valid piece!");
		}
		return null;
	}

	private static double getClosestPointRelativePostionOnSegment(Line2D line,
			int pointX, int pointY) {
		Point2D p1 = line.getOrigin();
		Point2D p2 = line.getEndPoint();
		double u = (((pointX - p1.getX()) * (p2.getX() - p1.getX())) + ((pointY - p1
				.getY()) * (p2.getY() - p1.getY())))
				/ (line.getDirection().getLength() * line.getDirection()
						.getLength());
		return u;
	}

	public void takeDone() {
		// TODO Auto-generated method stub

	}
}
